@@ -241,6 +241,12 @@ static int avalon_read(struct cgpu_info *avalon, unsigned char *buf,
 	char readbuf[AVALON_READBUF_SIZE];
 	int err, amount, ofs = 2, cp;
 
+	/* If the buffer is ready to take more work, yield once in case the
+	 * write thread is waiting to be scheduled. Keep it under the time
+	 * it would take to fill the entire 512 byte buffer. */
+	if (!avalon_buffer_full(avalon))
+		nmsleep(32);
+
 	err = usb_read_once_timeout(avalon, readbuf, readsize, &amount, timeout, ep);
 	applog(LOG_DEBUG, "%s%i: Get avalon read got err %d",
 	       avalon->drv->name, avalon->device_id, err);
@@ -331,7 +337,6 @@ static int avalon_reset(struct cgpu_info *avalon, bool initial)
 static bool get_options(int this_option_offset, int *baud, int *miner_count,
 			int *asic_count, int *timeout, int *frequency)
 {
-	char err_buf[BUFSIZ+1];
 	char buf[BUFSIZ+1];
 	char *ptr, *comma, *colon, *colon2, *colon3, *colon4;
 	size_t max;
@@ -382,10 +387,8 @@ static bool get_options(int this_option_offset, int *baud, int *miner_count,
 		*baud = 19200;
 		break;
 	default:
-		sprintf(err_buf,
-			"Invalid avalon-options for baud (%s) "
+		quit(1, "Invalid avalon-options for baud (%s) "
 			"must be 115200, 57600, 38400 or 19200", buf);
-		quit(1, err_buf);
 	}
 
 	if (colon && *colon) {
@@ -398,11 +401,9 @@ static bool get_options(int this_option_offset, int *baud, int *miner_count,
 			if (tmp > 0 && tmp <= AVALON_DEFAULT_MINER_NUM) {
 				*miner_count = tmp;
 			} else {
-				sprintf(err_buf,
-					"Invalid avalon-options for "
+				quit(1, "Invalid avalon-options for "
 					"miner_count (%s) must be 1 ~ %d",
 					colon, AVALON_DEFAULT_MINER_NUM);
-				quit(1, err_buf);
 			}
 		}
 
@@ -415,11 +416,9 @@ static bool get_options(int this_option_offset, int *baud, int *miner_count,
 			if (tmp > 0 && tmp <= AVALON_DEFAULT_ASIC_NUM)
 				*asic_count = tmp;
 			else {
-				sprintf(err_buf,
-					"Invalid avalon-options for "
+				quit(1, "Invalid avalon-options for "
 					"asic_count (%s) must be 1 ~ %d",
 					colon2, AVALON_DEFAULT_ASIC_NUM);
-				quit(1, err_buf);
 			}
 
 			if (colon3 && *colon3) {
@@ -431,11 +430,9 @@ static bool get_options(int this_option_offset, int *baud, int *miner_count,
 				if (tmp > 0 && tmp <= 0xff)
 					*timeout = tmp;
 				else {
-					sprintf(err_buf,
-						"Invalid avalon-options for "
+					quit(1, "Invalid avalon-options for "
 						"timeout (%s) must be 1 ~ %d",
 						colon3, 0xff);
-					quit(1, err_buf);
 				}
 				if (colon4 && *colon4) {
 					tmp = atoi(colon4);
@@ -447,10 +444,8 @@ static bool get_options(int this_option_offset, int *baud, int *miner_count,
 						*frequency = tmp;
 						break;
 					default:
-						sprintf(err_buf,
-							"Invalid avalon-options for "
+						quit(1, "Invalid avalon-options for "
 							"frequency must be 256/270/282/300");
-							quit(1, err_buf);
 					}
 				}
 			}
@@ -498,6 +493,16 @@ static void avalon_initialise(struct cgpu_info *avalon)
 	if (avalon->usbinfo.nodev)
 		return;
 
+	// Set latency
+	err = usb_transfer(avalon, FTDI_TYPE_OUT, FTDI_REQUEST_LATENCY,
+				AVALON_READ_TIMEOUT, interface, C_LATENCY);
+
+	applog(LOG_DEBUG, "%s%i: latency got err %d",
+		avalon->drv->name, avalon->device_id, err);
+
+	if (avalon->usbinfo.nodev)
+		return;
+
 	// Set data
 	err = usb_transfer(avalon, FTDI_TYPE_OUT, FTDI_REQUEST_DATA,
 				FTDI_VALUE_DATA_AVA, interface, C_SETDATA);
@@ -564,39 +569,27 @@ static bool avalon_detect_one(libusb_device *dev, struct usb_find_devices *found
 	int this_option_offset = ++option_offset;
 	struct avalon_info *info;
 	struct cgpu_info *avalon;
-	char devpath[20];
 	bool configured;
 	int ret;
 
-	avalon = calloc(1, sizeof(struct cgpu_info));
-	if (unlikely(!avalon))
-		quit(1, "Failed to calloc avalon in avalon_detect_one");;
-	avalon->drv = &avalon_drv;
-	avalon->threads = AVALON_MINER_THREADS;
+	avalon = usb_alloc_cgpu(&avalon_drv, AVALON_MINER_THREADS);
 
 	configured = get_options(this_option_offset, &baud, &miner_count,
 				 &asic_count, &timeout, &frequency);
 
 	if (!usb_init(avalon, dev, found))
-		return false;
+		goto shin;
 
 	/* Even though this is an FTDI type chip, we want to do the parsing
 	 * all ourselves so set it to std usb type */
 	avalon->usbdev->usb_type = USB_TYPE_STD;
 
 	/* We have a real Avalon! */
-	sprintf(devpath, "%d:%d",
-			(int)(avalon->usbinfo.bus_number),
-			(int)(avalon->usbinfo.device_address));
-
 	avalon_initialise(avalon);
 
-	avalon->device_path = strdup(devpath);
-	add_cgpu(avalon);
-
 	avalon->device_data = calloc(sizeof(struct avalon_info), 1);
 	if (unlikely(!(avalon->device_data)))
-		quit(1, "Failed to malloc avalon_info data");
+		quit(1, "Failed to calloc avalon_info data");
 	info = avalon->device_data;
 
 	if (configured) {
@@ -624,20 +617,36 @@ static bool avalon_detect_one(libusb_device *dev, struct usb_find_devices *found
 	info->temp_sum = 0;
 	info->temp_old = 0;
 
+	if (!add_cgpu(avalon))
+		goto unshin;
+
 	ret = avalon_reset(avalon, true);
-	if (ret && !configured) {
-		usb_uninit(avalon);
-		return false;
-	}
+	if (ret && !configured)
+		goto unshin;
+
+	update_usb_stats(avalon);
 
 	avalon_idle(avalon, info);
 
 	applog(LOG_DEBUG, "Avalon Detected: %s "
 	       "(miner_count=%d asic_count=%d timeout=%d frequency=%d)",
-	       devpath, info->miner_count, info->asic_count, info->timeout,
+	       avalon->device_path, info->miner_count, info->asic_count, info->timeout,
 	       info->frequency);
 
 	return true;
+
+unshin:
+
+	usb_uninit(avalon);
+
+shin:
+
+	free(avalon->device_data);
+	avalon->device_data = NULL;
+
+	avalon = usb_free_cgpu(avalon);
+
+	return false;
 }
 
 static void avalon_detect(void)
@@ -745,7 +754,6 @@ static void *avalon_get_results(void *userdata)
 	RenameThread(threadname);
 
 	while (likely(!avalon->shutdown)) {
-		struct timeval tv_start, now, tdiff;
 		unsigned char buf[rsize];
 		int ret;
 
@@ -760,27 +768,18 @@ static void *avalon_get_results(void *userdata)
 
 		if (unlikely(info->reset)) {
 			/* Tell the write thread it can start the reset */
-			sem_post(&info->write_sem);
-			sem_wait(&info->read_sem);
+			cgsem_post(&info->write_sem);
+			cgsem_wait(&info->read_sem);
 
 			/* Discard anything in the buffer */
 			offset = 0;
 		}
 
-		cgtime(&tv_start);
 		ret = avalon_read(avalon, buf, rsize, AVALON_READ_TIMEOUT,
 				  C_AVALON_READ);
 
-		if (ret < 1) {
-			int us_delay;
-
-			cgtime(&now);
-			timersub(&now, &tv_start, &tdiff);
-			us_delay = AVALON_READ_TIMEOUT * 1000 - (tdiff.tv_usec);
-			if (us_delay > 0)
-				nusleep(us_delay);
+		if (ret < 1)
 			continue;
-		}
 
 		if (opt_debug) {
 			applog(LOG_DEBUG, "Avalon: get:");
@@ -820,9 +819,9 @@ static void *avalon_send_tasks(void *userdata)
 		if (unlikely(info->reset)) {
 			/* Wait till read thread tells us it's received the
 			 * reset message */
-			sem_wait(&info->write_sem);
+			cgsem_wait(&info->write_sem);
 			avalon_running_reset(avalon, info);
-			sem_post(&info->read_sem);
+			cgsem_post(&info->read_sem);
 		}
 
 		mutex_lock(&info->qlock);
@@ -890,10 +889,8 @@ static bool avalon_prepare(struct thr_info *thr)
 	mutex_init(&info->qlock);
 	if (unlikely(pthread_cond_init(&info->qcond, NULL)))
 		quit(1, "Failed to pthread_cond_init avalon qcond");
-	if (unlikely(sem_init(&info->read_sem, 0, 0)))
-		quit(1, "Failed to sem_init avalon read_sem");
-	if (unlikely(sem_init(&info->write_sem, 0, 0)))
-		quit(1, "Failed to sem_init avalon write_sem");
+	cgsem_init(&info->read_sem);
+	cgsem_init(&info->write_sem);
 
 	if (pthread_create(&info->read_thr, NULL, avalon_get_results, (void *)avalon))
 		quit(1, "Failed to create avalon read_thr");
@@ -918,6 +915,9 @@ static void do_avalon_close(struct thr_info *thr)
 	avalon_running_reset(avalon, info);
 
 	info->no_matching_work = 0;
+
+	cgsem_destroy(&info->read_sem);
+	cgsem_destroy(&info->write_sem);
 }
 
 static inline void record_temp_fan(struct avalon_info *info, struct avalon_result *ar, float *temp_avg)
@@ -993,20 +993,15 @@ static void avalon_update_temps(struct cgpu_info *avalon, struct avalon_info *in
 static void get_avalon_statline_before(char *buf, struct cgpu_info *avalon)
 {
 	struct avalon_info *info = avalon->device_data;
-	int lowfan = 10000, pwm;
+	int lowfan = 10000;
 
-	/* Find the lowest fan speed. Fan0 is often not populated. */
-	if (info->fan0 > 0)
-		lowfan = info->fan0;
+	/* Find the lowest fan speed of the ASIC cooling fans. */
 	if (info->fan1 >= 0 && info->fan1 < lowfan)
 		lowfan = info->fan1;
 	if (info->fan2 >= 0 && info->fan2 < lowfan)
 		lowfan = info->fan2;
 
-	pwm = info->fan_pwm * 100 / AVALON_DEFAULT_FAN_MAX_PWM;
-
-	tailsprintf(buf, "%2d/%3dC %3d%%/%04dR| ", info->temp0, info->temp2,
-		    pwm, lowfan);
+	tailsprintf(buf, "%2d/%3dC %04dR | ", info->temp0, info->temp2, lowfan);
 }
 
 /* We use a replacement algorithm to only remove references to work done from
