@@ -13,7 +13,7 @@
 
 #ifdef USE_AVALON
 
-#include <semaphore.h>
+#include "util.h"
 
 #define AVALON_RESET_FAULT_DECISECONDS 1
 #define AVALON_MINER_THREADS 1
@@ -110,8 +110,8 @@ struct avalon_info {
 	pthread_mutex_t lock;
 	pthread_mutex_t qlock;
 	pthread_cond_t qcond;
-	sem_t read_sem;
-	sem_t write_sem;
+	cgsem_t read_sem;
+	cgsem_t write_sem;
 	int nonces;
 
 	bool idle;
