@@ -32,13 +32,16 @@
 
 /*
  * With Firmware 1.0.0 and a result queue of 20 the Max is:
- * header = 9
- * 64+1+32+1+1+(1+8)*8+1 per line = 172 * 20
+ * inprocess = 12
+ * max count = 9
+ * 64+1+24+1+1+(1+8)*8+1 per line = 164 * 20
  * OK = 3
- * Total: 3452
+ * Total: 3304
  */
 #define BFLSC_BUFSIZ (0x1000)
 
+#define BFLSC_INFO_TIMEOUT 999
+
 #define BFLSC_DI_FIRMWARE "FIRMWARE"
 #define BFLSC_DI_ENGINES "ENGINES"
 #define BFLSC_DI_JOBSINQUE "JOBS IN QUEUE"
@@ -113,6 +116,7 @@ struct bflsc_info {
 	bool shutdown;
 	bool flash_led;
 	bool not_first_work; // allow ignoring the first nonce error
+	bool fanauto;
 };
 
 #define BFLSC_XLINKHDR '@'
@@ -289,6 +293,11 @@ struct SaveString {
 #define BAJ_RES_TIME 100
 #define BFLSC_MAX_SLEEP 2000
 
+#define BAJ_LATENCY LATENCY_STD
+#define BAL_LATENCY LATENCY_STD
+#define BAS_LATENCY LATENCY_STD
+#define BAM_LATENCY 2
+
 #define BFLSC_TEMP_SLEEPMS 5
 #define BFLSC_QUE_SIZE 20
 #define BFLSC_QUE_FULL_ENOUGH 13
@@ -465,6 +474,12 @@ static int write_to_dev(struct cgpu_info *bflsc, int dev, char *buf, int buflen,
 	struct DataForwardToChain data;
 	int len;
 
+	/*
+	 * The protocol is syncronous so any previous excess can be
+	 * discarded and assumed corrupt data or failed USB transfers
+	 */
+	usb_buffer_clear(bflsc);
+
 	if (dev == 0)
 		return usb_write(bflsc, buf, buflen, amount, cmd);
 
@@ -576,6 +591,11 @@ static void __bflsc_initialise(struct cgpu_info *bflsc)
 	applog(LOG_DEBUG, "%s%i: reset got err %d",
 		bflsc->drv->name, bflsc->device_id, err);
 
+	if (bflsc->usbinfo.nodev)
+		return;
+
+	usb_ftdi_set_latency(bflsc);
+
 	if (bflsc->usbinfo.nodev)
 		return;
 
@@ -637,6 +657,8 @@ static void __bflsc_initialise(struct cgpu_info *bflsc)
 	applog(LOG_DEBUG, "%s%i: purgerx got err %d",
 		bflsc->drv->name, bflsc->device_id, err);
 
+	if (!bflsc->cutofftemp)
+		bflsc->cutofftemp = 80;
 }
 
 static void bflsc_initialise(struct cgpu_info *bflsc)
@@ -688,7 +710,8 @@ static bool getinfo(struct cgpu_info *bflsc, int dev)
 		return false;
 	}
 
-	err = usb_read_ok(bflsc, buf, sizeof(buf)-1, &amount, C_GETDETAILS);
+	err = usb_read_ok_timeout(bflsc, buf, sizeof(buf)-1, &amount,
+				  BFLSC_INFO_TIMEOUT, C_GETDETAILS);
 	if (err < 0 || amount < 1) {
 		if (err < 0) {
 			applog(LOG_ERR, "%s detect (%s) get details return invalid/timed out (%d:%d)",
@@ -779,20 +802,14 @@ static bool bflsc_detect_one(struct libusb_device *dev, struct usb_find_devices
 {
 	struct bflsc_info *sc_info = NULL;
 	char buf[BFLSC_BUFSIZ+1];
-	char devpath[20];
 	int i, err, amount;
 	struct timeval init_start, init_now;
 	int init_sleep, init_count;
 	bool ident_first;
 	char *newname;
+	uint16_t latency;
 
-	struct cgpu_info *bflsc = calloc(1, sizeof(*bflsc));
-
-	if (unlikely(!bflsc))
-		quit(1, "Failed to calloc bflsc in bflsc_detect_one");
-	bflsc->drv = &bflsc_drv;
-	bflsc->deven = DEV_ENABLED;
-	bflsc->threads = 1;
+	struct cgpu_info *bflsc = usb_alloc_cgpu(&bflsc_drv, 1);
 
 	sc_info = calloc(1, sizeof(*sc_info));
 	if (unlikely(!sc_info))
@@ -803,11 +820,6 @@ static bool bflsc_detect_one(struct libusb_device *dev, struct usb_find_devices
 	if (!usb_init(bflsc, dev, found))
 		goto shin;
 
-	sprintf(devpath, "%d:%d",
-			(int)(bflsc->usbinfo.bus_number),
-			(int)(bflsc->usbinfo.device_address));
-
-
 	// Allow 2 complete attempts if the 1st time returns an unrecognised reply
 	ident_first = true;
 retry:
@@ -819,7 +831,7 @@ reinit:
 	err = write_to_dev(bflsc, 0, BFLSC_IDENTIFY, BFLSC_IDENTIFY_LEN, &amount, C_REQUESTIDENTIFY);
 	if (err < 0 || amount != BFLSC_IDENTIFY_LEN) {
 		applog(LOG_ERR, "%s detect (%s) send identify request failed (%d:%d)",
-			bflsc->drv->dname, devpath, amount, err);
+			bflsc->drv->dname, bflsc->device_path, amount, err);
 		goto unshin;
 	}
 
@@ -830,7 +842,7 @@ reinit:
 		if (us_tdiff(&init_now, &init_start) <= REINIT_TIME_MAX) {
 			if (init_count == 2) {
 				applog(LOG_WARNING, "%s detect (%s) 2nd init failed (%d:%d) - retrying",
-					bflsc->drv->dname, devpath, amount, err);
+					bflsc->drv->dname, bflsc->device_path, amount, err);
 			}
 			nmsleep(init_sleep);
 			if ((init_sleep * 2) <= REINIT_TIME_MAX_MS)
@@ -840,14 +852,14 @@ reinit:
 
 		if (init_count > 0)
 			applog(LOG_WARNING, "%s detect (%s) init failed %d times %.2fs",
-				bflsc->drv->dname, devpath, init_count, tdiff(&init_now, &init_start));
+				bflsc->drv->dname, bflsc->device_path, init_count, tdiff(&init_now, &init_start));
 
 		if (err < 0) {
 			applog(LOG_ERR, "%s detect (%s) error identify reply (%d:%d)",
-				bflsc->drv->dname, devpath, amount, err);
+				bflsc->drv->dname, bflsc->device_path, amount, err);
 		} else {
 			applog(LOG_ERR, "%s detect (%s) empty identify reply (%d)",
-				bflsc->drv->dname, devpath, amount);
+				bflsc->drv->dname, bflsc->device_path, amount);
 		}
 
 		goto unshin;
@@ -856,30 +868,29 @@ reinit:
 
 	if (unlikely(!strstr(buf, BFLSC_BFLSC))) {
 		applog(LOG_DEBUG, "%s detect (%s) found an FPGA '%s' ignoring",
-			bflsc->drv->dname, devpath, buf);
+			bflsc->drv->dname, bflsc->device_path, buf);
 		goto unshin;
 	}
 
 	if (unlikely(strstr(buf, BFLSC_IDENTITY))) {
 		if (ident_first) {
 			applog(LOG_DEBUG, "%s detect (%s) didn't recognise '%s' trying again ...",
-				bflsc->drv->dname, devpath, buf);
+				bflsc->drv->dname, bflsc->device_path, buf);
 			ident_first = false;
 			goto retry;
 		}
 		applog(LOG_DEBUG, "%s detect (%s) didn't recognise '%s' on 2nd attempt",
-			bflsc->drv->dname, devpath, buf);
+			bflsc->drv->dname, bflsc->device_path, buf);
 		goto unshin;
 	}
 
-	bflsc->device_path = strdup(devpath);
-
 	if (!getinfo(bflsc, 0))
 		goto unshin;
 
 	sc_info->scan_sleep_time = BAS_SCAN_TIME;
 	sc_info->results_sleep_time = BAS_RES_TIME;
 	sc_info->default_ms_work = BAS_WORK_TIME;
+	latency = BAS_LATENCY;
 
 	/* When getinfo() "FREQUENCY: [UNKNOWN]" is fixed -
 	 * use 'freq * engines' to estimate.
@@ -891,6 +902,7 @@ reinit:
 		sc_info->results_sleep_time = BAM_RES_TIME;
 		sc_info->default_ms_work = BAM_WORK_TIME;
 		bflsc->usbdev->ident = IDENT_BAM;
+		latency = BAM_LATENCY;
 	} else {
 		if (sc_info->sc_devs[0].engines < 34) { // 16 * 2 + 2
 			newname = BFLSC_JALAPENO;
@@ -898,15 +910,22 @@ reinit:
 			sc_info->results_sleep_time = BAJ_RES_TIME;
 			sc_info->default_ms_work = BAJ_WORK_TIME;
 			bflsc->usbdev->ident = IDENT_BAJ;
+			latency = BAJ_LATENCY;
 		} else if (sc_info->sc_devs[0].engines < 130)  { // 16 * 8 + 2
 			newname = BFLSC_LITTLESINGLE;
 			sc_info->scan_sleep_time = BAL_SCAN_TIME;
 			sc_info->results_sleep_time = BAL_RES_TIME;
 			sc_info->default_ms_work = BAL_WORK_TIME;
 			bflsc->usbdev->ident = IDENT_BAL;
+			latency = BAL_LATENCY;
 		}
 	}
 
+	if (latency != bflsc->usbdev->found->latency) {
+		bflsc->usbdev->found->latency = latency;
+		usb_ftdi_set_latency(bflsc);
+	}
+
 	for (i = 0; i < sc_info->sc_count; i++)
 		sc_info->sc_devs[i].ms_work = sc_info->default_ms_work;
 
@@ -918,7 +937,7 @@ reinit:
 
 	// We have a real BFLSC!
 	applog(LOG_DEBUG, "%s (%s) identified as: '%s'",
-		bflsc->drv->dname, devpath, bflsc->drv->name);
+		bflsc->drv->dname, bflsc->device_path, bflsc->drv->name);
 
 	if (!add_cgpu(bflsc))
 		goto unshin;
@@ -928,6 +947,8 @@ reinit:
 	mutex_init(&bflsc->device_mutex);
 	rwlock_init(&sc_info->stat_lock);
 
+	usb_buffer_enable(bflsc);
+
 	return true;
 
 unshin:
@@ -936,16 +957,15 @@ unshin:
 
 shin:
 
-	free(bflsc->device_path);
 	free(bflsc->device_data);
+	bflsc->device_data = NULL;
 
-	if (bflsc->name != blank)
+	if (bflsc->name != blank) {
 		free(bflsc->name);
+		bflsc->name = NULL;
+	}
 
-	if (bflsc->drv->copy)
-		free(bflsc->drv);
-
-	free(bflsc);
+	bflsc = usb_free_cgpu(bflsc);
 
 	return false;
 }
@@ -1232,7 +1252,7 @@ static bool bflsc_get_temp(struct cgpu_info *bflsc, int dev)
 
 		bflsc->temp = temp;
 
-		if (bflsc->cutofftemp > 0 && temp > bflsc->cutofftemp) {
+		if (bflsc->cutofftemp > 0 && temp >= bflsc->cutofftemp) {
 			applog(LOG_WARNING, "%s%i:%s temp (%.1f) hit thermal cutoff limit %d, stopping work!",
 						bflsc->drv->name, bflsc->device_id, xlink,
 						temp, bflsc->cutofftemp);
@@ -1287,8 +1307,13 @@ static void process_nonces(struct cgpu_info *bflsc, int dev, char *xlink, char *
 
 	memset(midstate, 0, MIDSTATE_BYTES);
 	memset(blockdata, 0, MERKLE_BYTES);
-	hex2bin((unsigned char *)midstate, fields[QUE_MIDSTATE], MIDSTATE_BYTES);
-	hex2bin((unsigned char *)blockdata, fields[QUE_BLOCKDATA], MERKLE_BYTES);
+	if (!hex2bin((unsigned char *)midstate, fields[QUE_MIDSTATE], MIDSTATE_BYTES) ||
+	    !hex2bin((unsigned char *)blockdata, fields[QUE_BLOCKDATA], MERKLE_BYTES)) {
+		applog(LOG_ERR, "%s%i:%s Failed to convert binary data to hex result - ignored",
+		       bflsc->drv->name, bflsc->device_id, xlink);
+		inc_hw_errors(bflsc->thr[0]);
+		return;
+	}
 
 	work = find_queued_work_bymidstate(bflsc, midstate, MIDSTATE_BYTES,
 						blockdata, MERKLE_OFFSET, MERKLE_BYTES);
@@ -1767,6 +1792,29 @@ static int64_t bflsc_scanwork(struct thr_info *thr)
 	return ret;
 }
 
+static void bflsc_set_fanspeed(struct cgpu_info *bflsc)
+{
+	struct bflsc_info *sc_info = (struct bflsc_info *)bflsc->device_data;
+	int amount, err;
+
+	if ((bflsc->temp <= 60 && sc_info->fanauto) ||
+	    (bflsc->temp > 60 && !sc_info->fanauto))
+		return;
+
+	mutex_lock(&bflsc->device_mutex);
+	if (bflsc->temp > 60) {
+		write_to_dev(bflsc, 0, BFLSC_FAN4, BFLSC_FAN4_LEN, &amount,
+			     C_SETFAN);
+		sc_info->fanauto = false;
+	} else {
+		write_to_dev(bflsc, 0, BFLSC_FANAUTO, BFLSC_FANOUT_LEN,
+			     &amount, C_SETFAN);
+		sc_info->fanauto = true;
+	}
+	getok(bflsc, C_FANREPLY, &err, &amount);
+	mutex_unlock(&bflsc->device_mutex);
+}
+
 static bool bflsc_get_stats(struct cgpu_info *bflsc)
 {
 	struct bflsc_info *sc_info = (struct bflsc_info *)(bflsc->device_data);
@@ -1789,6 +1837,8 @@ static bool bflsc_get_stats(struct cgpu_info *bflsc)
 			nmsleep(BFLSC_TEMP_SLEEPMS);
 	}
 
+	bflsc_set_fanspeed(bflsc);
+
 	return allok;
 }
 
