@@ -256,16 +256,15 @@ static void icarus_initialise(struct cgpu_info *icarus, int baud)
 		case IDENT_LLT:
 		case IDENT_CMR1:
 		case IDENT_CMR2:
-			// Latency
-			transfer(icarus, FTDI_TYPE_OUT, FTDI_REQUEST_LATENCY, FTDI_VALUE_LATENCY,
-				 icarus->usbdev->found->interface, C_LATENCY);
+			// Reset
+			transfer(icarus, FTDI_TYPE_OUT, FTDI_REQUEST_RESET, FTDI_VALUE_RESET,
+				 icarus->usbdev->found->interface, C_RESET);
 
 			if (icarus->usbinfo.nodev)
 				return;
 
-			// Reset
-			transfer(icarus, FTDI_TYPE_OUT, FTDI_REQUEST_RESET, FTDI_VALUE_RESET,
-				 icarus->usbdev->found->interface, C_RESET);
+			// Latency
+			usb_ftdi_set_latency(icarus);
 
 			if (icarus->usbinfo.nodev)
 				return;
@@ -365,6 +364,16 @@ static void icarus_initialise(struct cgpu_info *icarus, int baud)
 		case IDENT_AMU:
 			nmsleep(20);
 
+			// Enable the UART
+			transfer(icarus, CP210X_TYPE_OUT, CP210X_REQUEST_IFC_ENABLE,
+				 CP210X_VALUE_UART_ENABLE,
+				 icarus->usbdev->found->interface, C_ENABLE_UART);
+
+			if (icarus->usbinfo.nodev)
+				return;
+
+			nmsleep(20);
+
 			// Set data control
 			transfer(icarus, CP210X_TYPE_OUT, CP210X_REQUEST_DATA, CP210X_VALUE_DATA,
 				 icarus->usbdev->found->interface, C_SETDATA);
@@ -433,11 +442,10 @@ static int icarus_get_nonce(struct cgpu_info *icarus, unsigned char *buf, struct
 		if (first)
 			copy_time(tv_finish, &read_finish);
 
-		// TODO: test if there is more data? to read a 2nd nonce?
 		if (amt >= read_amount)
 			return ICA_NONCE_OK;
 
-		rc += SECTOMS(tdiff(&read_finish, &read_start));
+		rc = SECTOMS(tdiff(&read_finish, tv_start));
 		if (rc >= read_time) {
 			if (amt > 0)
 				applog(LOG_DEBUG, "Icarus Read: Timeout reading for %d ms", rc);
@@ -587,7 +595,6 @@ static void set_timing_mode(int this_option_offset, struct cgpu_info *icarus)
 
 static uint32_t mask(int work_division)
 {
-	char err_buf[BUFSIZ+1];
 	uint32_t nonce_mask = 0x7fffffff;
 
 	// yes we can calculate these, but this way it's easy to see what they are
@@ -605,8 +612,7 @@ static uint32_t mask(int work_division)
 		nonce_mask = 0x1fffffff;
 		break;
 	default:
-		sprintf(err_buf, "Invalid2 icarus-options for work_division (%d) must be 1, 2, 4 or 8", work_division);
-		quit(1, err_buf);
+		quit(1, "Invalid2 icarus-options for work_division (%d) must be 1, 2, 4 or 8", work_division);
 	}
 
 	return nonce_mask;
@@ -614,7 +620,6 @@ static uint32_t mask(int work_division)
 
 static void get_options(int this_option_offset, struct cgpu_info *icarus, int *baud, int *work_division, int *fpga_count)
 {
-	char err_buf[BUFSIZ+1];
 	char buf[BUFSIZ+1];
 	char *ptr, *comma, *colon, *colon2;
 	size_t max;
@@ -683,8 +688,7 @@ static void get_options(int this_option_offset, struct cgpu_info *icarus, int *b
 				*baud = 57600;
 				break;
 			default:
-				sprintf(err_buf, "Invalid icarus-options for baud (%s) must be 115200 or 57600", buf);
-				quit(1, err_buf);
+				quit(1, "Invalid icarus-options for baud (%s) must be 115200 or 57600", buf);
 			}
 		}
 
@@ -699,8 +703,7 @@ static void get_options(int this_option_offset, struct cgpu_info *icarus, int *b
 					*work_division = tmp;
 					*fpga_count = tmp;	// default to the same
 				} else {
-					sprintf(err_buf, "Invalid icarus-options for work_division (%s) must be 1, 2, 4 or 8", colon);
-					quit(1, err_buf);
+					quit(1, "Invalid icarus-options for work_division (%s) must be 1, 2, 4 or 8", colon);
 				}
 			}
 
@@ -709,8 +712,7 @@ static void get_options(int this_option_offset, struct cgpu_info *icarus, int *b
 				if (tmp > 0 && tmp <= *work_division)
 					*fpga_count = tmp;
 				else {
-					sprintf(err_buf, "Invalid icarus-options for fpga_count (%s) must be >0 and <=work_division (%d)", colon2, *work_division);
-					quit(1, err_buf);
+					quit(1, "Invalid icarus-options for fpga_count (%s) must be >0 and <=work_division (%d)", colon2, *work_division);
 				}
 			}
 		}
@@ -720,7 +722,6 @@ static void get_options(int this_option_offset, struct cgpu_info *icarus, int *b
 static bool icarus_detect_one(struct libusb_device *dev, struct usb_find_devices *found)
 {
 	int this_option_offset = ++option_offset;
-	char devpath[20];
 	struct ICARUS_INFO *info;
 	struct timeval tv_start, tv_finish;
 
@@ -743,24 +744,13 @@ static bool icarus_detect_one(struct libusb_device *dev, struct usb_find_devices
 	int ret, err, amount, tries;
 	bool ok;
 
-	icarus = calloc(1, sizeof(struct cgpu_info));
-	if (unlikely(!icarus))
-		quit(1, "Failed to calloc icarus in icarus_detect_one");
-	icarus->drv = &icarus_drv;
-	icarus->deven = DEV_ENABLED;
-	icarus->threads = 1;
+	icarus = usb_alloc_cgpu(&icarus_drv, 1);
 
 	if (!usb_init(icarus, dev, found))
 		goto shin;
 
 	get_options(this_option_offset, icarus, &baud, &work_division, &fpga_count);
 
-	sprintf(devpath, "%d:%d",
-			(int)(icarus->usbinfo.bus_number),
-			(int)(icarus->usbinfo.device_address));
-
-	icarus->device_path = strdup(devpath);
-
 	hex2bin(ob_bin, golden_ob, sizeof(ob_bin));
 
 	tries = 2;
@@ -786,7 +776,7 @@ static bool icarus_detect_one(struct libusb_device *dev, struct usb_find_devices
 				applog(LOG_ERR,
 					"Icarus Detect: "
 					"Test failed at %s: get %s, should: %s",
-					devpath, nonce_hex, golden_nonce);
+					icarus->device_path, nonce_hex, golden_nonce);
 			}
 		}
 		free(nonce_hex);
@@ -798,7 +788,7 @@ static bool icarus_detect_one(struct libusb_device *dev, struct usb_find_devices
 	applog(LOG_DEBUG,
 		"Icarus Detect: "
 		"Test succeeded at %s: got %s",
-			devpath, golden_nonce);
+			icarus->device_path, golden_nonce);
 
 	/* We have a real Icarus! */
 	if (!add_cgpu(icarus))
@@ -807,7 +797,7 @@ static bool icarus_detect_one(struct libusb_device *dev, struct usb_find_devices
 	update_usb_stats(icarus);
 
 	applog(LOG_INFO, "%s%d: Found at %s",
-		icarus->drv->name, icarus->device_id, devpath);
+		icarus->drv->name, icarus->device_id, icarus->device_path);
 
 	applog(LOG_DEBUG, "%s%d: Init baud=%d work_division=%d fpga_count=%d",
 		icarus->drv->name, icarus->device_id, baud, work_division, fpga_count);
@@ -831,17 +821,17 @@ static bool icarus_detect_one(struct libusb_device *dev, struct usb_find_devices
 
 	set_timing_mode(this_option_offset, icarus);
 
+	usb_buffer_enable(icarus);
+
 	return true;
 
 unshin:
 
 	usb_uninit(icarus);
 
-	free(icarus->device_path);
-
 shin:
 
-	free(icarus);
+	icarus = usb_free_cgpu(icarus);
 
 	return false;
 }
@@ -899,6 +889,9 @@ static int64_t icarus_scanhash(struct thr_info *thr, struct work *work,
 	rev(ob_bin, 32);
 	rev(ob_bin + 52, 12);
 
+	// We only want results for the work we are about to send
+	usb_buffer_clear(icarus);
+
 	err = usb_write(icarus, (char *)ob_bin, sizeof(ob_bin), &amount, C_SENDWORK);
 	if (err < 0 || amount != sizeof(ob_bin)) {
 		applog(LOG_ERR, "%s%i: Comms error (werr=%d amt=%d)",
@@ -957,6 +950,20 @@ static int64_t icarus_scanhash(struct thr_info *thr, struct work *work,
 	hash_count++;
 	hash_count *= info->fpga_count;
 
+#if 0
+	// This appears to only return zero nonce values
+	if (usb_buffer_size(icarus) > 3) {
+		memcpy((char *)&nonce, icarus->usbdev->buffer, sizeof(nonce_bin));
+		nonce = htobe32(nonce);
+		applog(LOG_WARNING, "%s%d: attempting to submit 2nd nonce = 0x%08lX",
+				icarus->drv->name, icarus->device_id,
+				(long unsigned int)nonce);
+		curr_hw_errors = icarus->hw_errors;
+		submit_nonce(thr, work, nonce);
+		was_hw_error = (curr_hw_errors > icarus->hw_errors);
+	}
+#endif
+
 	if (opt_debug || info->do_icarus_timing)
 		timersub(&tv_finish, &tv_start, &elapsed);
 
