@@ -44,7 +44,6 @@ extern void enable_curses(void);
 extern int mining_threads;
 extern double total_secs;
 extern int opt_g_threads;
-extern bool ping;
 extern bool opt_loginput;
 extern char *opt_kernel_path;
 extern int gpur_thr_id;
@@ -54,7 +53,6 @@ extern bool have_opencl;
 extern void *miner_thread(void *userdata);
 extern int dev_from_id(int thr_id);
 extern void tailsprintf(char *f, const char *fmt, ...);
-extern void wlog(const char *f, ...);
 extern void decay_time(double *f, double fadd);
 
 /**********************************************/
@@ -229,7 +227,7 @@ char *set_kernel(char *arg)
 	char *nextptr;
 
 	if (opt_scrypt)
-		return "Cannot use sha256 kernel with scrypt";
+		return "Cannot specify a kernel with scrypt";
 	nextptr = strtok(arg, ",");
 	if (nextptr == NULL)
 		return "Invalid parameters for set kernel";
@@ -285,14 +283,6 @@ char *set_gpu_map(char *arg)
 	return NULL;
 }
 
-void get_intrange(char *arg, int *val1, int *val2)
-{
-	if (sscanf(arg, "%d-%d", val1, val2) == 1) {
-		*val2 = *val1;
-		*val1 = 0;
-	}
-}
-
 char *set_gpu_engine(char *arg)
 {
 	int i, val1 = 0, val2 = 0, device = 0;
@@ -626,7 +616,7 @@ void pause_dynamic_threads(int gpu)
 
 		thr->pause = cgpu->dynamic;
 		if (!cgpu->dynamic && cgpu->deven != DEV_DISABLED)
-			tq_push(thr->q, &ping);
+			cgsem_post(&thr->sem);
 	}
 }
 
@@ -695,7 +685,7 @@ retry:
 				if (powertune != -1)
 					tailsprintf(logline, "P: %d%%", powertune);
 				tailsprintf(logline, "\n");
-				wlog(logline);
+				_wlog(logline);
 			}
 		}
 #endif
@@ -772,9 +762,9 @@ retry:
 				wlogprint("Must restart device before enabling it");
 				goto retry;
 			}
-			applog(LOG_DEBUG, "Pushing ping to thread %d", thr->id);
+			applog(LOG_DEBUG, "Pushing sem post to thread %d", thr->id);
 
-			tq_push(thr->q, &ping);
+			cgsem_post(&thr->sem);
 		}
 		goto retry;
 	} if (!strncasecmp(&input, "d", 1)) {
@@ -1218,7 +1208,7 @@ select_cgpu:
 		if (dev_from_id(thr_id) != gpu)
 			continue;
 
-		tq_push(thr->q, &ping);
+		cgsem_post(&thr->sem);
 	}
 
 	goto select_cgpu;
