@@ -22,6 +22,10 @@
 			(err) == LIBUSB_ERROR_PIPE || \
 			(err) == LIBUSB_ERROR_OTHER)
 
+#define NOCONTROLDEV(err) ((err) == LIBUSB_ERROR_NO_DEVICE || \
+			(err) == LIBUSB_ERROR_OTHER)
+
+			
 #ifdef USE_BFLSC
 #define DRV_BFLSC 1
 #endif
@@ -64,6 +68,8 @@
 #define ICARUS_TIMEOUT_MS 200
 #endif
 
+#define USB_READ_MINPOLL 40
+
 #ifdef USE_BFLSC
 // N.B. transfer size is 512 with USB2.0, but only 64 with USB1.1
 static struct usb_endpoints bas_eps[] = {
@@ -147,6 +153,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = BFLSC_TIMEOUT_MS,
+		.latency = LATENCY_STD,
 		.epcount = ARRAY_SIZE(bas_eps),
 		.eps = bas_eps },
 #endif
@@ -163,6 +170,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = BITFORCE_TIMEOUT_MS,
+		.latency = LATENCY_STD,
 		.epcount = ARRAY_SIZE(bfl_eps),
 		.eps = bfl_eps },
 #endif
@@ -177,6 +185,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 1,
 		.timeout = MODMINER_TIMEOUT_MS,
+		.latency = LATENCY_UNUSED,
 		.epcount = ARRAY_SIZE(mmq_eps),
 		.eps = mmq_eps },
 #endif
@@ -191,6 +200,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = AVALON_TIMEOUT_MS,
+		.latency = 10,
 		.epcount = ARRAY_SIZE(ava_eps),
 		.eps = ava_eps },
 #endif
@@ -205,6 +215,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = ICARUS_TIMEOUT_MS,
+		.latency = LATENCY_UNUSED,
 		.epcount = ARRAY_SIZE(ica_eps),
 		.eps = ica_eps },
 	{
@@ -217,6 +228,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = ICARUS_TIMEOUT_MS,
+		.latency = LATENCY_UNUSED,
 		.epcount = ARRAY_SIZE(amu_eps),
 		.eps = amu_eps },
 	{
@@ -230,6 +242,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = ICARUS_TIMEOUT_MS,
+		.latency = LATENCY_STD,
 		.epcount = ARRAY_SIZE(llt_eps),
 		.eps = llt_eps },
 	// For any that don't match the above "BLT"
@@ -243,6 +256,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = ICARUS_TIMEOUT_MS,
+		.latency = LATENCY_STD,
 		.epcount = ARRAY_SIZE(llt_eps),
 		.eps = llt_eps },
 	{
@@ -256,6 +270,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = ICARUS_TIMEOUT_MS,
+		.latency = LATENCY_STD,
 		.epcount = ARRAY_SIZE(cmr1_eps),
 		.eps = cmr1_eps },
 	{
@@ -269,6 +284,7 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 0,
 		.timeout = ICARUS_TIMEOUT_MS,
+		.latency = LATENCY_STD,
 		.epcount = ARRAY_SIZE(cmr2_eps),
 		.eps = cmr2_eps },
 #endif
@@ -285,10 +301,11 @@ static struct usb_find_devices find_dev[] = {
 		.config = 1,
 		.interface = 1,
 		.timeout = 100,
+		.latency = LATENCY_UNUSED,
 		.epcount = 0,
 		.eps = NULL },
 #endif
-	{ DRV_LAST, NULL, 0, 0, 0, NULL, NULL, 0, 0, 0, 0, 0, NULL }
+	{ DRV_LAST, NULL, 0, 0, 0, NULL, NULL, 0, 0, 0, 0, 0, 0, NULL }
 };
 
 #ifdef USE_BFLSC
@@ -438,7 +455,7 @@ struct cg_usb_stats {
 #define SEQ1 1
 
 static struct cg_usb_stats *usb_stats = NULL;
-static int next_stat = 0;
+static int next_stat = USB_NOSTAT;
 
 #define USB_STATS(sgpu, sta, fin, err, mode, cmd, seq) \
 		stats(cgpu, sta, fin, err, mode, cmd, seq)
@@ -504,12 +521,15 @@ static const char *C_SENDTESTWORK_S = "SendTestWork";
 static const char *C_LATENCY_S = "SetLatency";
 static const char *C_SETLINE_S = "SetLine";
 static const char *C_VENDOR_S = "Vendor";
+static const char *C_SETFAN_S = "SetFan";
+static const char *C_FANREPLY_S = "GetFan";
 static const char *C_AVALON_TASK_S = "AvalonTask";
 static const char *C_AVALON_READ_S = "AvalonRead";
 static const char *C_GET_AVALON_READY_S = "AvalonReady";
 static const char *C_AVALON_RESET_S = "AvalonReset";
 static const char *C_GET_AVALON_RESET_S = "GetAvalonReset";
 static const char *C_FTDI_STATUS_S = "FTDIStatus";
+static const char *C_ENABLE_UART_S = "EnableUART";
 
 #ifdef EOL
 #undef EOL
@@ -904,7 +924,7 @@ void usb_all(int level)
 		for (i = 0; i < count; i++)
 			usb_full(&j, list[i], &buf, &off, &len, level);
 
-		applog(LOG_WARNING, "%s", buf);
+		_applog(LOG_WARNING, buf);
 
 		free(buf);
 
@@ -988,12 +1008,15 @@ static void cgusb_check_init()
 		usb_commands[C_LATENCY] = C_LATENCY_S;
 		usb_commands[C_SETLINE] = C_SETLINE_S;
 		usb_commands[C_VENDOR] = C_VENDOR_S;
+		usb_commands[C_SETFAN] = C_SETFAN_S;
+		usb_commands[C_FANREPLY] = C_FANREPLY_S;
 		usb_commands[C_AVALON_TASK] = C_AVALON_TASK_S;
 		usb_commands[C_AVALON_READ] = C_AVALON_READ_S;
 		usb_commands[C_GET_AVALON_READY] = C_GET_AVALON_READY_S;
 		usb_commands[C_AVALON_RESET] = C_AVALON_RESET_S;
 		usb_commands[C_GET_AVALON_RESET] = C_GET_AVALON_RESET_S;
 		usb_commands[C_FTDI_STATUS] = C_FTDI_STATUS_S;
+		usb_commands[C_ENABLE_UART] = C_ENABLE_UART_S;
 
 		stats_initialised = true;
 	}
@@ -1212,7 +1235,7 @@ static bool cgminer_usb_lock_bd(struct device_drv *drv, uint8_t bus_number, uint
 	res_work_head = res_work;
 	mutex_unlock(&cgusbres_lock);
 
-	sem_post(&usb_resource_sem);
+	cgsem_post(&usb_resource_sem);
 
 	// TODO: add a timeout fail - restart the resource thread?
 	while (true) {
@@ -1271,7 +1294,7 @@ static void cgminer_usb_unlock_bd(struct device_drv *drv, uint8_t bus_number, ui
 	res_work_head = res_work;
 	mutex_unlock(&cgusbres_lock);
 
-	sem_post(&usb_resource_sem);
+	cgsem_post(&usb_resource_sem);
 
 	return;
 }
@@ -1298,6 +1321,9 @@ static struct cg_usb_device *free_cgusb(struct cg_usb_device *cgusb)
 
 	free(cgusb->found);
 
+	if (cgusb->buffer)
+		free(cgusb->buffer);
+
 	free(cgusb);
 
 	return NULL;
@@ -1312,11 +1338,18 @@ void usb_uninit(struct cgpu_info *cgpu)
 	//  if release_cgpu() was called due to a USB NODEV(err)
 	if (!cgpu->usbdev)
 		return;
-	libusb_release_interface(cgpu->usbdev->handle, cgpu->usbdev->found->interface);
-	libusb_close(cgpu->usbdev->handle);
+	if (!libusb_release_interface(cgpu->usbdev->handle, cgpu->usbdev->found->interface)) {
+		cg_wlock(&cgusb_fd_lock);
+		libusb_close(cgpu->usbdev->handle);
+		cg_wunlock(&cgusb_fd_lock);
+	}
 	cgpu->usbdev = free_cgusb(cgpu->usbdev);
 }
 
+/*
+ * N.B. this is always called inside
+ *	DEVLOCK(cgpu, pstate);
+ */
 static void release_cgpu(struct cgpu_info *cgpu)
 {
 	struct cg_usb_device *cgusb = cgpu->usbdev;
@@ -1330,6 +1363,7 @@ static void release_cgpu(struct cgpu_info *cgpu)
 	if (cgpu->usbinfo.nodev)
 		return;
 
+	zombie_devs++;
 	total_count--;
 	drv_count[cgpu->drv->drv_id].count--;
 
@@ -1358,10 +1392,88 @@ static void release_cgpu(struct cgpu_info *cgpu)
 	cgminer_usb_unlock_bd(cgpu->drv, cgpu->usbinfo.bus_number, cgpu->usbinfo.device_address);
 }
 
+// Currently only used by MMQ
+struct cgpu_info *usb_copy_cgpu(struct cgpu_info *orig)
+{
+	struct cgpu_info *copy = calloc(1, sizeof(*copy));
+
+	if (unlikely(!copy))
+		quit(1, "Failed to calloc cgpu for %s in usb_copy_cgpu", orig->drv->dname);
+
+	copy->name = orig->name;
+	copy->drv = copy_drv(orig->drv);
+	copy->deven = orig->deven;
+	copy->threads = orig->threads;
+
+	copy->usbdev = orig->usbdev;
+
+	memcpy(&(copy->usbinfo), &(orig->usbinfo), sizeof(copy->usbinfo));
+
+	copy->usbinfo.nodev = (copy->usbdev == NULL);
+
+	copy->usbinfo.devlock = orig->usbinfo.devlock;
+
+	return copy;
+}
+
+struct cgpu_info *usb_alloc_cgpu(struct device_drv *drv, int threads)
+{
+	struct cgpu_info *cgpu = calloc(1, sizeof(*cgpu));
+
+	if (unlikely(!cgpu))
+		quit(1, "Failed to calloc cgpu for %s in usb_alloc_cgpu", drv->dname);
+
+	cgpu->drv = drv;
+	cgpu->deven = DEV_ENABLED;
+	cgpu->threads = threads;
+
+	cgpu->usbinfo.nodev = true;
+
+	cgpu->usbinfo.devlock = calloc(1, sizeof(*(cgpu->usbinfo.devlock)));
+	if (unlikely(!cgpu->usbinfo.devlock))
+		quit(1, "Failed to calloc devlock for %s in usb_alloc_cgpu", drv->dname);
+
+	rwlock_init(cgpu->usbinfo.devlock);
+
+	return cgpu;
+}
+
+struct cgpu_info *usb_free_cgpu_devlock(struct cgpu_info *cgpu, bool free_devlock)
+{
+	if (cgpu->drv->copy)
+		free(cgpu->drv);
+
+	free(cgpu->device_path);
+
+	if (free_devlock)
+		free(cgpu->usbinfo.devlock);
+
+	free(cgpu);
+
+	return NULL;
+}
+
 #define USB_INIT_FAIL 0
 #define USB_INIT_OK 1
 #define USB_INIT_IGNORE 2
 
+/*
+ * WARNING - these assume DEVLOCK(cgpu, pstate) is called first and
+ *  DEVUNLOCK(cgpu, pstate) in called in the same function with the same pstate
+ *  given to DEVLOCK.
+ *  You must call DEVUNLOCK(cgpu, pstate) before exiting the function or it will leave
+ *  the thread Cancelability unrestored
+ */
+#define DEVLOCK(cgpu, _pth_state) do { \
+			pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &_pth_state); \
+			wr_lock(cgpu->usbinfo.devlock); \
+			} while (0)
+
+#define DEVUNLOCK(cgpu, _pth_state) do { \
+			wr_unlock(cgpu->usbinfo.devlock); \
+			pthread_setcanceltype(_pth_state, NULL); \
+			} while (0)
+
 static int _usb_init(struct cgpu_info *cgpu, struct libusb_device *dev, struct usb_find_devices *found)
 {
 	struct cg_usb_device *cgusb = NULL;
@@ -1369,15 +1481,23 @@ static int _usb_init(struct cgpu_info *cgpu, struct libusb_device *dev, struct u
 	const struct libusb_interface_descriptor *idesc;
 	const struct libusb_endpoint_descriptor *epdesc;
 	unsigned char strbuf[STRBUFLEN+1];
+	char devpath[32];
 	char devstr[STRBUFLEN+1];
-	int err, i, j, k;
+	int err, i, j, k, pstate;
 	int bad = USB_INIT_FAIL;
 
+	DEVLOCK(cgpu, pstate);
+
 	cgpu->usbinfo.bus_number = libusb_get_bus_number(dev);
 	cgpu->usbinfo.device_address = libusb_get_device_address(dev);
 
-	sprintf(devstr, "- %s device %d:%d", found->name,
-		cgpu->usbinfo.bus_number, cgpu->usbinfo.device_address);
+	sprintf(devpath, "%d:%d",
+		(int)(cgpu->usbinfo.bus_number),
+		(int)(cgpu->usbinfo.device_address));
+
+	cgpu->device_path = strdup(devpath);
+
+	sprintf(devstr, "- %s device %s", found->name, devpath);
 
 	cgusb = calloc(1, sizeof(*cgusb));
 	if (unlikely(!cgusb))
@@ -1401,7 +1521,9 @@ static int _usb_init(struct cgpu_info *cgpu, struct libusb_device *dev, struct u
 		goto dame;
 	}
 
+	cg_wlock(&cgusb_fd_lock);
 	err = libusb_open(dev, &(cgusb->handle));
+	cg_wunlock(&cgusb_fd_lock);
 	if (err) {
 		switch (err) {
 			case LIBUSB_ERROR_ACCESS:
@@ -1415,7 +1537,7 @@ static int _usb_init(struct cgpu_info *cgpu, struct libusb_device *dev, struct u
 			case LIBUSB_ERROR_NOT_SUPPORTED:
 				applog(LOG_ERR,
 					"USB init, open device failed, err %d, "
-					"you need to install a Windows USB driver for %s",
+					"you need to install a WinUSB driver for %s",
 					err, devstr);
 				break;
 #endif
@@ -1594,6 +1716,7 @@ static int _usb_init(struct cgpu_info *cgpu, struct libusb_device *dev, struct u
 		cgusb->manuf_string, cgusb->serial_string);
 
 	cgpu->usbdev = cgusb;
+	cgpu->usbinfo.nodev = false;
 
 	libusb_free_config_descriptor(config);
 
@@ -1605,11 +1728,14 @@ static int _usb_init(struct cgpu_info *cgpu, struct libusb_device *dev, struct u
 		cgpu->drv->name = (char *)(found->name);
 	}
 
-	return USB_INIT_OK;
+	bad = USB_INIT_OK;
+	goto out_unlock;
 
 cldame:
 
+	cg_wlock(&cgusb_fd_lock);
 	libusb_close(cgusb->handle);
+	cg_wunlock(&cgusb_fd_lock);
 
 dame:
 
@@ -1618,6 +1744,9 @@ dame:
 
 	cgusb = free_cgusb(cgusb);
 
+out_unlock:
+	DEVUNLOCK(cgpu, pstate);
+
 	return bad;
 }
 
@@ -1761,7 +1890,7 @@ void usb_detect(struct device_drv *drv, bool (*device_detect)(struct libusb_devi
 
 	count = libusb_get_device_list(NULL, &list);
 	if (count < 0) {
-		applog(LOG_DEBUG, "USB scan devices: failed, err %zd", count);
+		applog(LOG_DEBUG, "USB scan devices: failed, err %d", (int)count);
 		return;
 	}
 
@@ -1854,7 +1983,7 @@ struct api_data *api_usb_stats(__maybe_unused int *count)
 	int cmdseq;
 	char modes_s[32];
 
-	if (next_stat == 0)
+	if (next_stat == USB_NOSTAT)
 		return NULL;
 
 	while (*count < next_stat * C_MAX * 2) {
@@ -1927,21 +2056,25 @@ static void newstats(struct cgpu_info *cgpu)
 	int i;
 
 	mutex_lock(&cgusb_lock);
-	cgpu->usbinfo.usbstat = ++next_stat;
-	mutex_unlock(&cgusb_lock);
 
-	usb_stats = realloc(usb_stats, sizeof(*usb_stats) * next_stat);
+	cgpu->usbinfo.usbstat = next_stat + 1;
+
+	usb_stats = realloc(usb_stats, sizeof(*usb_stats) * (next_stat+1));
 	if (unlikely(!usb_stats))
-		quit(1, "USB failed to realloc usb_stats %d", next_stat);
+		quit(1, "USB failed to realloc usb_stats %d", next_stat+1);
 
-	usb_stats[next_stat-1].name = cgpu->drv->name;
-	usb_stats[next_stat-1].device_id = -1;
-	usb_stats[next_stat-1].details = calloc(1, sizeof(struct cg_usb_stats_details) * C_MAX * 2);
-	if (unlikely(!usb_stats[next_stat-1].details))
-		quit(1, "USB failed to calloc details for %d", next_stat);
+	usb_stats[next_stat].name = cgpu->drv->name;
+	usb_stats[next_stat].device_id = -1;
+	usb_stats[next_stat].details = calloc(1, sizeof(struct cg_usb_stats_details) * C_MAX * 2);
+	if (unlikely(!usb_stats[next_stat].details))
+		quit(1, "USB failed to calloc details for %d", next_stat+1);
 
 	for (i = 1; i < C_MAX * 2; i += 2)
-		usb_stats[next_stat-1].details[i].seq = 1;
+		usb_stats[next_stat].details[i].seq = 1;
+
+	next_stat++;
+
+	mutex_unlock(&cgusb_lock);
 }
 #endif
 
@@ -2011,11 +2144,86 @@ static void rejected_inc(struct cgpu_info *cgpu, uint32_t mode)
 }
 #endif
 
+static char *find_end(unsigned char *buf, unsigned char *ptr, int ptrlen, int tot, char *end, int endlen, bool first)
+{
+	unsigned char *search;
+
+	if (endlen > tot)
+		return NULL;
+
+	// If end is only 1 char - do a faster search
+	if (endlen == 1) {
+		if (first)
+			search = buf;
+		else
+			search = ptr;
+
+		return strchr((char *)search, *end);
+	} else {
+		if (first)
+			search = buf;
+		else {
+			// must allow end to have been chopped in 2
+			if ((tot - ptrlen) >= (endlen - 1))
+				search = ptr - (endlen - 1);
+			else
+				search = ptr - (tot - ptrlen);
+		}
+
+		return strstr((char *)search, end);
+	}
+}
+
 #define USB_MAX_READ 8192
+#define USB_RETRY_MAX 5
+
+static int
+usb_bulk_transfer(struct libusb_device_handle *dev_handle,
+		  unsigned char endpoint, unsigned char *data, int length,
+		  int *transferred, unsigned int timeout,
+		  struct cgpu_info *cgpu)
+{
+	int err, tries = 0;
+
+	cg_rlock(&cgusb_fd_lock);
+	err = libusb_bulk_transfer(dev_handle, endpoint, data, length,
+				   transferred, timeout);
+	cg_runlock(&cgusb_fd_lock);
+
+	if (err == LIBUSB_ERROR_PIPE) {
+		cgpu->usbinfo.last_pipe = time(NULL);
+		cgpu->usbinfo.pipe_count++;
+		applog(LOG_INFO, "%s%i: libusb pipe error, trying to clear",
+			cgpu->drv->name, cgpu->device_id);
+		do {
+			err = libusb_clear_halt(dev_handle, endpoint);
+			if (unlikely(err == LIBUSB_ERROR_NOT_FOUND ||
+				     err == LIBUSB_ERROR_NO_DEVICE)) {
+					cgpu->usbinfo.clear_err_count++;
+					break;
+			}
+
+			cg_rlock(&cgusb_fd_lock);
+			err = libusb_bulk_transfer(dev_handle, endpoint, data,
+						   length, transferred, timeout);
+			cg_runlock(&cgusb_fd_lock);
+
+			if (err)
+				cgpu->usbinfo.retry_err_count++;
+		} while (err == LIBUSB_ERROR_PIPE && tries++ < USB_RETRY_MAX);
+		applog(LOG_DEBUG, "%s%i: libusb pipe error%scleared",
+			cgpu->drv->name, cgpu->device_id, err ? " not " : " ");
+
+		if (err)
+			cgpu->usbinfo.clear_fail_count++;
+	}
+
+	return err;
+}
 
 int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *processed, unsigned int timeout, const char *end, __maybe_unused enum usb_cmds cmd, bool readonce)
 {
-	struct cg_usb_device *usbdev = cgpu->usbdev;
+	struct cg_usb_device *usbdev;
 	bool ftdi;
 #if DO_USB_STATS
 	struct timeval tv_start;
@@ -2023,51 +2231,71 @@ int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pro
 	struct timeval read_start, tv_finish;
 	unsigned int initial_timeout;
 	double max, done;
-	int bufleft, err, got, tot;
-	__maybe_unused bool first = true;
-	unsigned char *search;
+	int bufleft, err, got, tot, pstate;
+	bool first = true;
+	char *search;
 	int endlen;
 
 	// We add 4: 1 for null, 2 for FTDI status and 1 to round to 4 bytes
 	unsigned char usbbuf[USB_MAX_READ+4], *ptr;
 	size_t usbbufread;
 
+	DEVLOCK(cgpu, pstate);
+
 	if (cgpu->usbinfo.nodev) {
 		*buf = '\0';
 		*processed = 0;
 		USB_REJECT(cgpu, MODE_BULK_READ);
 
-		return LIBUSB_ERROR_NO_DEVICE;
+		err = LIBUSB_ERROR_NO_DEVICE;
+		goto out_unlock;
 	}
 
+	usbdev = cgpu->usbdev;
 	ftdi = (usbdev->usb_type == USB_TYPE_FTDI);
 
 	USBDEBUG("USB debug: _usb_read(%s (nodev=%s),ep=%d,buf=%p,bufsiz=%zu,proc=%p,timeout=%u,end=%s,cmd=%s,ftdi=%s,readonce=%s)", cgpu->drv->name, bool_str(cgpu->usbinfo.nodev), ep, buf, bufsiz, processed, timeout, end ? (char *)str_text((char *)end) : "NULL", usb_cmdname(cmd), bool_str(ftdi), bool_str(readonce));
 
 	if (bufsiz > USB_MAX_READ)
-		quit(1, "%s USB read request %d too large (max=%d)", cgpu->drv->name, bufsiz, USB_MAX_READ);
+		quit(1, "%s USB read request %d too large (max=%d)", cgpu->drv->name, (int)bufsiz, USB_MAX_READ);
 
 	if (timeout == DEVTIMEOUT)
 		timeout = usbdev->found->timeout;
 
 	if (end == NULL) {
-		tot = 0;
-		ptr = usbbuf;
-		bufleft = bufsiz;
+		if (usbdev->buffer && usbdev->bufamt) {
+			tot = usbdev->bufamt;
+			bufleft = bufsiz - tot;
+			memcpy(usbbuf, usbdev->buffer, tot);
+			ptr = usbbuf + tot;
+			usbdev->bufamt = 0;
+		} else {
+			tot = 0;
+			bufleft = bufsiz;
+			ptr = usbbuf;
+		}
+
 		err = LIBUSB_SUCCESS;
 		initial_timeout = timeout;
 		max = ((double)timeout) / 1000.0;
 		cgtime(&read_start);
 		while (bufleft > 0) {
-			if (ftdi)
-				usbbufread = bufleft + 2;
-			else
-				usbbufread = bufleft;
+			// TODO: use (USB_MAX_READ - tot) always?
+			if (usbdev->buffer)
+				usbbufread = USB_MAX_READ - tot;
+			else {
+				if (ftdi)
+					usbbufread = bufleft + 2;
+				else
+					usbbufread = bufleft;
+			}
 			got = 0;
+
 			STATS_TIMEVAL(&tv_start);
-			err = libusb_bulk_transfer(usbdev->handle,
-					usbdev->found->eps[ep].ep,
-					ptr, usbbufread, &got, timeout);
+			err = usb_bulk_transfer(usbdev->handle,
+						usbdev->found->eps[ep].ep,
+						ptr, usbbufread, &got, timeout,
+						cgpu);
 			cgtime(&tv_finish);
 			USB_STATS(cgpu, &tv_start, &tv_finish, err,
 					MODE_BULK_READ, cmd, first ? SEQ0 : SEQ1);
@@ -2102,8 +2330,19 @@ int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pro
 			// N.B. this is: return LIBUSB_SUCCESS with whatever size has already been read
 			if (unlikely(done >= max))
 				break;
-
 			timeout = initial_timeout - (done * 1000);
+			if (!timeout)
+				break;
+		}
+
+		// N.B. usbdev->buffer was emptied before the while() loop
+		if (usbdev->buffer && tot > (int)bufsiz) {
+			usbdev->bufamt = tot - bufsiz;
+			memcpy(usbdev->buffer, ptr + bufsiz, usbdev->bufamt);
+			tot -= usbdev->bufamt;
+			usbbuf[tot] = '\0';
+			applog(LOG_ERR, "USB: %s%i read1 buffering %d extra bytes",
+					cgpu->drv->name, cgpu->device_id, usbdev->bufamt);
 		}
 
 		*processed = tot;
@@ -2112,27 +2351,41 @@ int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pro
 		if (NODEV(err))
 			release_cgpu(cgpu);
 
-		return err;
+		goto out_unlock;
+	}
+
+	if (usbdev->buffer && usbdev->bufamt) {
+		tot = usbdev->bufamt;
+		bufleft = bufsiz - tot;
+		memcpy(usbbuf, usbdev->buffer, tot);
+		ptr = usbbuf + tot;
+		usbdev->bufamt = 0;
+	} else {
+		tot = 0;
+		bufleft = bufsiz;
+		ptr = usbbuf;
 	}
 
-	tot = 0;
-	ptr = usbbuf;
-	bufleft = bufsiz;
 	endlen = strlen(end);
 	err = LIBUSB_SUCCESS;
 	initial_timeout = timeout;
 	max = ((double)timeout) / 1000.0;
 	cgtime(&read_start);
 	while (bufleft > 0) {
-		if (ftdi)
-			usbbufread = bufleft + 2;
-		else
-			usbbufread = bufleft;
+		// TODO: use (USB_MAX_READ - tot) always?
+		if (usbdev->buffer)
+			usbbufread = USB_MAX_READ - tot;
+		else {
+			if (ftdi)
+				usbbufread = bufleft + 2;
+			else
+				usbbufread = bufleft;
+		}
 		got = 0;
 		STATS_TIMEVAL(&tv_start);
-		err = libusb_bulk_transfer(usbdev->handle,
-				usbdev->found->eps[ep].ep,
-				ptr, usbbufread, &got, timeout);
+		err = usb_bulk_transfer(usbdev->handle,
+					usbdev->found->eps[ep].ep, ptr,
+					usbbufread, &got, timeout, cgpu);
 		cgtime(&tv_finish);
 		USB_STATS(cgpu, &tv_start, &tv_finish, err,
 				MODE_BULK_READ, cmd, first ? SEQ0 : SEQ1);
@@ -2158,23 +2411,8 @@ int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pro
 		if (err || readonce)
 			break;
 
-		// WARNING - this will return data past END ('if' there is extra data)
-		if (endlen <= tot) {
-			// If END is only 1 char - do a faster search
-			if (endlen == 1) {
-				if (strchr((char *)ptr, *end))
-					break;
-			} else {
-				// must allow END to have been chopped in 2 transfers
-				if ((tot - got) >= (endlen - 1))
-					search = ptr - (endlen - 1);
-				else
-					search = ptr - (tot - got);
-
-				if (strstr((char *)search, end))
-					break;
-			}
-		}
+		if (find_end(usbbuf, ptr, got, tot, (char *)end, endlen, first))
+			break;
 
 		ptr += got;
 		bufleft -= got;
@@ -2185,8 +2423,41 @@ int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pro
 		// N.B. this is: return LIBUSB_SUCCESS with whatever size has already been read
 		if (unlikely(done >= max))
 			break;
-
 		timeout = initial_timeout - (done * 1000);
+		if (!timeout)
+			break;
+	}
+
+	if (usbdev->buffer) {
+		bool dobuffer = false;
+
+		if ((search = find_end(usbbuf, usbbuf, tot, tot, (char *)end, endlen, true))) {
+			// end finishes after bufsiz
+			if ((search + endlen - (char *)usbbuf) > (int)bufsiz) {
+				usbdev->bufamt = tot - bufsiz;
+				dobuffer = true;
+			} else {
+				// extra data after end
+				if (*(search + endlen)) {
+					usbdev->bufamt = tot - (search + endlen - (char *)usbbuf);
+					dobuffer = true;
+				}
+			}
+		} else {
+			// no end, but still bigger than bufsiz
+			if (tot > (int)bufsiz) {
+				usbdev->bufamt = tot - bufsiz;
+				dobuffer = true;
+			}
+		}
+
+		if (dobuffer) {
+			tot -= usbdev->bufamt;
+			memcpy(usbdev->buffer, usbbuf + tot, usbdev->bufamt);
+			usbbuf[tot] = '\0';
+			applog(LOG_ERR, "USB: %s%i read2 buffering %d extra bytes",
+					cgpu->drv->name, cgpu->device_id, usbdev->bufamt);
+		}
 	}
 
 	*processed = tot;
@@ -2195,12 +2466,15 @@ int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pro
 	if (NODEV(err))
 		release_cgpu(cgpu);
 
+out_unlock:
+	DEVUNLOCK(cgpu, pstate);
+
 	return err;
 }
 
 int _usb_write(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *processed, unsigned int timeout, __maybe_unused enum usb_cmds cmd)
 {
-	struct cg_usb_device *usbdev = cgpu->usbdev;
+	struct cg_usb_device *usbdev;
 #if DO_USB_STATS
 	struct timeval tv_start;
 #endif
@@ -2208,7 +2482,9 @@ int _usb_write(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pr
 	unsigned int initial_timeout;
 	double max, done;
 	__maybe_unused bool first = true;
-	int err, sent, tot;
+	int err, sent, tot, pstate;
+
+	DEVLOCK(cgpu, pstate);
 
 	USBDEBUG("USB debug: _usb_write(%s (nodev=%s),ep=%d,buf='%s',bufsiz=%zu,proc=%p,timeout=%u,cmd=%s)", cgpu->drv->name, bool_str(cgpu->usbinfo.nodev), ep, (char *)str_text(buf), bufsiz, processed, timeout, usb_cmdname(cmd));
 
@@ -2217,9 +2493,12 @@ int _usb_write(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pr
 	if (cgpu->usbinfo.nodev) {
 		USB_REJECT(cgpu, MODE_BULK_WRITE);
 
-		return LIBUSB_ERROR_NO_DEVICE;
+		err = LIBUSB_ERROR_NO_DEVICE;
+		goto out_unlock;
 	}
 
+	usbdev = cgpu->usbdev;
+
 	if (timeout == DEVTIMEOUT)
 		timeout = usbdev->found->timeout;
 
@@ -2231,10 +2510,10 @@ int _usb_write(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pr
 	while (bufsiz > 0) {
 		sent = 0;
 		STATS_TIMEVAL(&tv_start);
-		err = libusb_bulk_transfer(usbdev->handle,
-				usbdev->found->eps[ep].ep,
-				(unsigned char *)buf,
-				bufsiz, &sent, timeout);
+		err = usb_bulk_transfer(usbdev->handle,
+					usbdev->found->eps[ep].ep,
+					(unsigned char *)buf, bufsiz, &sent,
+					timeout, cgpu);
 		cgtime(&tv_finish);
 		USB_STATS(cgpu, &tv_start, &tv_finish, err,
 				MODE_BULK_WRITE, cmd, first ? SEQ0 : SEQ1);
@@ -2257,8 +2536,9 @@ int _usb_write(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pr
 		// N.B. this is: return LIBUSB_SUCCESS with whatever size was written
 		if (unlikely(done >= max))
 			break;
-
 		timeout = initial_timeout - (done * 1000);
+		if (!timeout)
+			break;
 	}
 
 	*processed = tot;
@@ -2266,12 +2546,15 @@ int _usb_write(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *pr
 	if (NODEV(err))
 		release_cgpu(cgpu);
 
+out_unlock:
+	DEVUNLOCK(cgpu, pstate);
+
 	return err;
 }
 
-int _usb_transfer(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex, uint32_t *data, int siz, unsigned int timeout, __maybe_unused enum usb_cmds cmd)
+int __usb_transfer(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex, uint32_t *data, int siz, unsigned int timeout, __maybe_unused enum usb_cmds cmd)
 {
-	struct cg_usb_device *usbdev = cgpu->usbdev;
+	struct cg_usb_device *usbdev;
 #if DO_USB_STATS
 	struct timeval tv_start, tv_finish;
 #endif
@@ -2283,8 +2566,10 @@ int _usb_transfer(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest
 	if (cgpu->usbinfo.nodev) {
 		USB_REJECT(cgpu, MODE_CTRL_WRITE);
 
-		return LIBUSB_ERROR_NO_DEVICE;
+		err = LIBUSB_ERROR_NO_DEVICE;
+		goto out_;
 	}
+	usbdev = cgpu->usbdev;
 
 	USBDEBUG("USB debug: @_usb_transfer() data=%s", bin2hex((unsigned char *)data, (size_t)siz));
 
@@ -2302,9 +2587,11 @@ int _usb_transfer(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest
 	USBDEBUG("USB debug: @_usb_transfer() buf=%s", bin2hex((unsigned char *)buf, (size_t)siz));
 
 	STATS_TIMEVAL(&tv_start);
+	cg_rlock(&cgusb_fd_lock);
 	err = libusb_control_transfer(usbdev->handle, request_type,
 		bRequest, wValue, wIndex, (unsigned char *)buf, (uint16_t)siz,
 		timeout == DEVTIMEOUT ? usbdev->found->timeout : timeout);
+	cg_runlock(&cgusb_fd_lock);
 	STATS_TIMEVAL(&tv_finish);
 	USB_STATS(cgpu, &tv_start, &tv_finish, err, MODE_CTRL_WRITE, cmd, SEQ0);
 
@@ -2312,38 +2599,57 @@ int _usb_transfer(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest
 
 	IOERR_CHECK(cgpu, err);
 
-	if (buf)
-		free(buf);
+	free(buf);
 
-	if (NODEV(err))
+	if (NOCONTROLDEV(err))
 		release_cgpu(cgpu);
 
+out_:
+	return err;
+}
+
+int _usb_transfer(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex, uint32_t *data, int siz, unsigned int timeout, enum usb_cmds cmd)
+{
+	int pstate, err;
+
+	DEVLOCK(cgpu, pstate);
+
+	err = __usb_transfer(cgpu, request_type, bRequest, wValue, wIndex, data, siz, timeout, cmd);
+
+	DEVUNLOCK(cgpu, pstate);
+
 	return err;
 }
 
 int _usb_transfer_read(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex, char *buf, int bufsiz, int *amount, unsigned int timeout, __maybe_unused enum usb_cmds cmd)
 {
-	struct cg_usb_device *usbdev = cgpu->usbdev;
+	struct cg_usb_device *usbdev;
 #if DO_USB_STATS
 	struct timeval tv_start, tv_finish;
 #endif
-	int err;
+	int err, pstate;
+
+	DEVLOCK(cgpu, pstate);
 
 	USBDEBUG("USB debug: _usb_transfer_read(%s (nodev=%s),type=%"PRIu8",req=%"PRIu8",value=%"PRIu16",index=%"PRIu16",bufsiz=%d,timeout=%u,cmd=%s)", cgpu->drv->name, bool_str(cgpu->usbinfo.nodev), request_type, bRequest, wValue, wIndex, bufsiz, timeout, usb_cmdname(cmd));
 
 	if (cgpu->usbinfo.nodev) {
 		USB_REJECT(cgpu, MODE_CTRL_READ);
 
-		return LIBUSB_ERROR_NO_DEVICE;
+		err = LIBUSB_ERROR_NO_DEVICE;
+		goto out_unlock;
 	}
+	usbdev = cgpu->usbdev;
 
 	*amount = 0;
 
 	STATS_TIMEVAL(&tv_start);
+	cg_rlock(&cgusb_fd_lock);
 	err = libusb_control_transfer(usbdev->handle, request_type,
 		bRequest, wValue, wIndex,
 		(unsigned char *)buf, (uint16_t)bufsiz,
 		timeout == DEVTIMEOUT ? usbdev->found->timeout : timeout);
+	cg_runlock(&cgusb_fd_lock);
 	STATS_TIMEVAL(&tv_finish);
 	USB_STATS(cgpu, &tv_start, &tv_finish, err, MODE_CTRL_READ, cmd, SEQ0);
 
@@ -2354,9 +2660,12 @@ int _usb_transfer_read(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRe
 	if (err > 0) {
 		*amount = err;
 		err = 0;
-	} else if (NODEV(err))
+	} else if (NOCONTROLDEV(err))
 		release_cgpu(cgpu);
 
+out_unlock:
+	DEVUNLOCK(cgpu, pstate);
+
 	return err;
 }
 
@@ -2384,6 +2693,133 @@ int usb_ftdi_cts(struct cgpu_info *cgpu)
 	return (ret & FTDI_RS0_CTS);
 }
 
+int usb_ftdi_set_latency(struct cgpu_info *cgpu)
+{
+	int err = 0;
+	int pstate;
+
+	DEVLOCK(cgpu, pstate);
+
+	if (cgpu->usbdev) {
+		if (cgpu->usbdev->usb_type != USB_TYPE_FTDI) {
+			applog(LOG_ERR, "%s: cgid %d latency request on non-FTDI device",
+				cgpu->drv->name, cgpu->cgminer_id);
+			err = LIBUSB_ERROR_NOT_SUPPORTED;
+		} else if (cgpu->usbdev->found->latency == LATENCY_UNUSED) {
+			applog(LOG_ERR, "%s: cgid %d invalid latency (UNUSED)",
+				cgpu->drv->name, cgpu->cgminer_id);
+			err = LIBUSB_ERROR_NOT_SUPPORTED;
+		}
+
+		if (!err)
+			err = __usb_transfer(cgpu, FTDI_TYPE_OUT, FTDI_REQUEST_LATENCY,
+						cgpu->usbdev->found->latency,
+						cgpu->usbdev->found->interface,
+						NULL, 0, DEVTIMEOUT, C_LATENCY);
+	}
+
+	DEVUNLOCK(cgpu, pstate);
+
+	applog(LOG_DEBUG, "%s: cgid %d %s got err %d",
+				cgpu->drv->name, cgpu->cgminer_id,
+				usb_cmdname(C_LATENCY), err);
+
+	return err;
+}
+
+void usb_buffer_enable(struct cgpu_info *cgpu)
+{
+	struct cg_usb_device *cgusb;
+	int pstate;
+
+	DEVLOCK(cgpu, pstate);
+
+	cgusb = cgpu->usbdev;
+	if (cgusb && !cgusb->buffer) {
+		cgusb->bufamt = 0;
+		cgusb->buffer = malloc(USB_MAX_READ+1);
+		if (!cgusb->buffer)
+			quit(1, "Failed to malloc buffer for USB %s%i",
+				cgpu->drv->name, cgpu->device_id);
+		cgusb->bufsiz = USB_MAX_READ;
+	}
+
+	DEVUNLOCK(cgpu, pstate);
+}
+
+void usb_buffer_disable(struct cgpu_info *cgpu)
+{
+	struct cg_usb_device *cgusb;
+	int pstate;
+
+	DEVLOCK(cgpu, pstate);
+
+	cgusb = cgpu->usbdev;
+	if (cgusb && cgusb->buffer) {
+		cgusb->bufamt = 0;
+		cgusb->bufsiz = 0;
+		free(cgusb->buffer);
+		cgusb->buffer = NULL;
+	}
+
+	DEVUNLOCK(cgpu, pstate);
+}
+
+void usb_buffer_clear(struct cgpu_info *cgpu)
+{
+	int pstate;
+
+	DEVLOCK(cgpu, pstate);
+
+	if (cgpu->usbdev)
+		cgpu->usbdev->bufamt = 0;
+
+	DEVUNLOCK(cgpu, pstate);
+}
+
+uint32_t usb_buffer_size(struct cgpu_info *cgpu)
+{
+	uint32_t ret = 0;
+	int pstate;
+
+	DEVLOCK(cgpu, pstate);
+
+	if (cgpu->usbdev)
+		ret = cgpu->usbdev->bufamt;
+
+	DEVUNLOCK(cgpu, pstate);
+
+	return ret;
+}
+
+// Need to set all devices with matching usbdev
+void usb_set_dev_start(struct cgpu_info *cgpu)
+{
+	struct cg_usb_device *cgusb;
+	struct cgpu_info *cgpu2;
+	struct timeval now;
+	int pstate;
+
+	DEVLOCK(cgpu, pstate);
+
+	cgusb = cgpu->usbdev;
+
+	// If the device wasn't dropped
+	if (cgusb != NULL) {
+		int i;
+
+		cgtime(&now);
+
+		for (i = 0; i < total_devices; i++) {
+			cgpu2 = get_devices(i);
+			if (cgpu2->usbdev == cgusb)
+				copy_time(&(cgpu2->dev_start_tv), &now);
+		}
+	}
+
+	DEVUNLOCK(cgpu, pstate);
+}
+
 void usb_cleanup()
 {
 	struct cgpu_info *cgpu;
@@ -2403,7 +2839,9 @@ void usb_cleanup()
 			case DRIVER_MODMINER:
 			case DRIVER_ICARUS:
 			case DRIVER_AVALON:
+				wr_lock(cgpu->usbinfo.devlock);
 				release_cgpu(cgpu);
+				wr_unlock(cgpu->usbinfo.devlock);
 				count++;
 				break;
 			default:
@@ -2438,6 +2876,8 @@ void usb_cleanup()
 		}
 		mutex_unlock(&cgusbres_lock);
 	}
+
+	cgsem_destroy(&usb_resource_sem);
 }
 
 void usb_initialise()
@@ -2569,15 +3009,18 @@ void usb_initialise()
 #include <sys/types.h>
 #include <sys/ipc.h>
 #include <sys/sem.h>
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 
+#ifndef __APPLE__
 union semun {
-	int sem;
-	struct semid_ds *seminfo;
-	ushort *all;
+	int val;
+	struct semid_ds *buf;
+	unsigned short *array;
+	struct seminfo *__buf;
 };
+#endif
+
 #else
 static LPSECURITY_ATTRIBUTES unsec(LPSECURITY_ATTRIBUTES sec)
 {
@@ -2602,7 +3045,7 @@ static LPSECURITY_ATTRIBUTES mksec(const char *dname, uint8_t bus_number, uint8_
 		applog(LOG_ERR,
 			"MTX: %s (%d:%d) USB failed to init secdes err (%d)",
 			dname, (int)bus_number, (int)device_address,
-			GetLastError());
+			(int)GetLastError());
 		free(sec_des);
 		return NULL;
 	}
@@ -2611,7 +3054,7 @@ static LPSECURITY_ATTRIBUTES mksec(const char *dname, uint8_t bus_number, uint8_
 		applog(LOG_ERR,
 			"MTX: %s (%d:%d) USB failed to secdes dacl err (%d)",
 			dname, (int)bus_number, (int)device_address,
-			GetLastError());
+			(int)GetLastError());
 		free(sec_des);
 		return NULL;
 	}
@@ -2620,7 +3063,7 @@ static LPSECURITY_ATTRIBUTES mksec(const char *dname, uint8_t bus_number, uint8_
 		applog(LOG_ERR,
 			"MTX: %s (%d:%d) USB failed to create gsid err (%d)",
 			dname, (int)bus_number, (int)device_address,
-			GetLastError());
+			(int)GetLastError());
 		free(sec_des);
 		return NULL;
 	}
@@ -2629,7 +3072,7 @@ static LPSECURITY_ATTRIBUTES mksec(const char *dname, uint8_t bus_number, uint8_
 		applog(LOG_ERR,
 			"MTX: %s (%d:%d) USB failed to secdes grp err (%d)",
 			dname, (int)bus_number, (int)device_address,
-			GetLastError());
+			(int)GetLastError());
 		FreeSid(gsid);
 		free(sec_des);
 		return NULL;
@@ -2674,7 +3117,7 @@ static bool resource_lock(const char *dname, uint8_t bus_number, uint8_t device_
 	if (usbMutex == NULL) {
 		applog(LOG_ERR,
 			"MTX: %s USB failed to get '%s' err (%d)",
-			dname, name, GetLastError());
+			dname, name, (int)GetLastError());
 		sec = unsec(sec);
 		return false;
 	}
@@ -2698,7 +3141,7 @@ static bool resource_lock(const char *dname, uint8_t bus_number, uint8_t device_
 					}
 					applog(LOG_ERR,
 						"MTX: %s USB can't get '%s' - device in use - failure (%d)",
-						dname, name, GetLastError());
+						dname, name, (int)GetLastError());
 					goto fail;
 				}
 			}
@@ -2712,12 +3155,12 @@ static bool resource_lock(const char *dname, uint8_t bus_number, uint8_t device_
 		case WAIT_FAILED:
 			applog(LOG_ERR,
 				"MTX: %s USB failed to get '%s' err (%d)",
-				dname, name, GetLastError());
+				dname, name, (int)GetLastError());
 			goto fail;
 		default:
 			applog(LOG_ERR,
 				"MTX: %s USB failed to get '%s' unknown reply (%d)",
-				dname, name, res);
+				dname, name, (int)res);
 			goto fail;
 	}
 
@@ -2766,7 +3209,7 @@ fail:
 			return false;
 		}
 
-		opt.seminfo = &seminfo;
+		opt.buf = &seminfo;
 		count = 0;
 		while (++count) {
 			// Should NEVER take 100ms
@@ -2782,7 +3225,7 @@ fail:
 					dname, sem, name, count, errno, strerror(errno));
 				return false;
 			}
-			if (opt.seminfo->sem_otime != 0)
+			if (opt.buf->sem_otime != 0)
 				break;
 			nmsleep(1);
 		}
@@ -2833,7 +3276,7 @@ static void resource_unlock(const char *dname, uint8_t bus_number, uint8_t devic
 	if (!ReleaseMutex(usbMutex))
 		applog(LOG_ERR,
 			"MTX: %s USB failed to release '%s' err (%d)",
-			dname, name, GetLastError());
+			dname, name, (int)GetLastError());
 
 fila:
 
@@ -2940,7 +3383,7 @@ void *usb_resource_thread(void __maybe_unused *userdata)
 
 	while (42) {
 		/* Wait to be told we have work to do */
-		sem_wait(&usb_resource_sem);
+		cgsem_wait(&usb_resource_sem);
 
 		mutex_lock(&cgusbres_lock);
 		while (res_work_head)
