@@ -68,9 +68,11 @@
 // For 0x10c4:0xea60 USB cp210x chip - AMU
 #define CP210X_TYPE_OUT 0x41
 
+#define CP210X_REQUEST_IFC_ENABLE 0x00
 #define CP210X_REQUEST_DATA 0x07
 #define CP210X_REQUEST_BAUD 0x1e
 
+#define CP210X_VALUE_UART_ENABLE 0x0001
 #define CP210X_VALUE_DATA 0x0303
 #define CP210X_DATA_BAUD 0x0001c200
 
@@ -140,10 +142,17 @@ struct usb_find_devices {
 	int config;
 	int interface;
 	unsigned int timeout;
+	uint16_t latency;
 	int epcount;
 	struct usb_endpoints *eps;
 };
 
+/* Latency is set to 32ms to prevent a transfer ever being more than 512 bytes
+ * +2 bytes of status such as the ftdi chip, when the chips emulate a 115200
+ * baud rate, to avoid status bytes being interleaved in larger transfers. */
+#define LATENCY_UNUSED 0
+#define LATENCY_STD 32
+
 enum usb_types {
 	USB_TYPE_STD = 0,
 	USB_TYPE_FTDI
@@ -163,8 +172,13 @@ struct cg_usb_device {
 	char *serial_string;
 	unsigned char fwVersion;	// ??
 	unsigned char interfaceVersion;	// ??
+	char *buffer;
+	uint32_t bufsiz;
+	uint32_t bufamt;
 };
 
+#define USB_NOSTAT 0
+
 struct cg_usb_info {
 	uint8_t bus_number;
 	uint8_t device_address;
@@ -174,6 +188,23 @@ struct cg_usb_info {
 	struct timeval last_nodev;
 	uint32_t ioerr_count;
 	uint32_t continuous_ioerr_count;
+
+	/*
+	 * for nodev and cgusb access (read and write)
+	 * it's a pointer so MMQ can have it in multiple devices
+	 *
+	 * N.B. general mining code doesn't need to use the read
+	 * lock for 'nodev' if it calls a usb_read/write/etc function
+	 * that uses the lock - however, all usbutils code MUST use it
+	 * to avoid devices disappearing while in use by multiple threads
+	 */
+	pthread_rwlock_t *devlock;
+
+	time_t last_pipe;
+	uint64_t pipe_count;
+	uint64_t clear_err_count;
+	uint64_t retry_err_count;
+	uint64_t clear_fail_count;
 };
 
 enum usb_cmds {
@@ -228,12 +259,15 @@ enum usb_cmds {
 	C_LATENCY,
 	C_SETLINE,
 	C_VENDOR,
+	C_SETFAN,
+	C_FANREPLY,
 	C_AVALON_TASK,
 	C_AVALON_READ,
 	C_GET_AVALON_READY,
 	C_AVALON_RESET,
 	C_GET_AVALON_RESET,
 	C_FTDI_STATUS,
+	C_ENABLE_UART,
 	C_MAX
 };
 
@@ -243,17 +277,26 @@ struct cgpu_info;
 void usb_all(int level);
 const char *usb_cmdname(enum usb_cmds cmd);
 void usb_applog(struct cgpu_info *bflsc, enum usb_cmds cmd, char *msg, int amount, int err);
+struct cgpu_info *usb_copy_cgpu(struct cgpu_info *orig);
+struct cgpu_info *usb_alloc_cgpu(struct device_drv *drv, int threads);
+struct cgpu_info *usb_free_cgpu_devlock(struct cgpu_info *cgpu, bool free_devlock);
+#define usb_free_cgpu(cgpu) usb_free_cgpu_devlock(cgpu, true)
 void usb_uninit(struct cgpu_info *cgpu);
 bool usb_init(struct cgpu_info *cgpu, struct libusb_device *dev, struct usb_find_devices *found);
 void usb_detect(struct device_drv *drv, bool (*device_detect)(struct libusb_device *, struct usb_find_devices *));
 struct api_data *api_usb_stats(int *count);
 void update_usb_stats(struct cgpu_info *cgpu);
 int _usb_read(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *processed, unsigned int timeout, const char *end, enum usb_cmds cmd, bool readonce);
-int usb_ftdi_cts(struct cgpu_info *cgpu);
-int usb_ftdi_ctw(struct cgpu_info *cgpu);
 int _usb_write(struct cgpu_info *cgpu, int ep, char *buf, size_t bufsiz, int *processed, unsigned int timeout, enum usb_cmds);
 int _usb_transfer(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex, uint32_t *data, int siz, unsigned int timeout, enum usb_cmds cmd);
 int _usb_transfer_read(struct cgpu_info *cgpu, uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex, char *buf, int bufsiz, int *amount, unsigned int timeout, enum usb_cmds cmd);
+int usb_ftdi_cts(struct cgpu_info *cgpu);
+int usb_ftdi_set_latency(struct cgpu_info *cgpu);
+void usb_buffer_enable(struct cgpu_info *cgpu);
+void usb_buffer_disable(struct cgpu_info *cgpu);
+void usb_buffer_clear(struct cgpu_info *cgpu);
+uint32_t usb_buffer_size(struct cgpu_info *cgpu);
+void usb_set_dev_start(struct cgpu_info *cgpu);
 void usb_cleanup();
 void usb_initialise();
 void *usb_resource_thread(void *userdata);
@@ -273,6 +316,9 @@ void *usb_resource_thread(void *userdata);
 #define usb_read_ok(cgpu, buf, bufsiz, read, cmd) \
 	_usb_read(cgpu, DEFAULT_EP_IN, buf, bufsiz, read, DEVTIMEOUT, "OK\n", cmd, false)
 
+#define usb_read_ok_timeout(cgpu, buf, bufsiz, read, timeout, cmd) \
+	_usb_read(cgpu, DEFAULT_EP_IN, buf, bufsiz, read, timeout, "OK\n", cmd, false)
+
 #define usb_read_ep(cgpu, ep, buf, bufsiz, read, cmd) \
 	_usb_read(cgpu, ep, buf, bufsiz, read, DEVTIMEOUT, NULL, cmd, false)
 
