@@ -615,7 +615,7 @@ bool hex2bin(unsigned char *p, const char *hexstr, size_t len)
 		hex_byte[1] = hexstr[1];
 
 		if (unlikely(sscanf(hex_byte, "%x", &v) != 1)) {
-			applog(LOG_ERR, "hex2bin sscanf '%s' failed", hex_byte);
+			applog(LOG_INFO, "hex2bin sscanf '%s' failed", hex_byte);
 			return ret;
 		}
 
@@ -785,30 +785,11 @@ out:
 
 int thr_info_create(struct thr_info *thr, pthread_attr_t *attr, void *(*start) (void *), void *arg)
 {
+	cgsem_init(&thr->sem);
+
 	return pthread_create(&thr->pth, attr, start, arg);
 }
 
-void thr_info_freeze(struct thr_info *thr)
-{
-	struct tq_ent *ent, *iter;
-	struct thread_q *tq;
-
-	if (!thr)
-		return;
-
-	tq = thr->q;
-	if (!tq)
-		return;
-
-	mutex_lock(&tq->mutex);
-	tq->frozen = true;
-	list_for_each_entry_safe(ent, iter, &tq->q, q_node) {
-		list_del(&ent->q_node);
-		free(ent);
-	}
-	mutex_unlock(&tq->mutex);
-}
-
 void thr_info_cancel(struct thr_info *thr)
 {
 	if (!thr)
@@ -818,6 +799,7 @@ void thr_info_cancel(struct thr_info *thr)
 		pthread_cancel(thr->pth);
 		PTH(thr) = 0L;
 	}
+	cgsem_destroy(&thr->sem);
 }
 
 /* Provide a ms based sleep that uses nanosleep to avoid poor usleep accuracy
@@ -1030,9 +1012,11 @@ bool stratum_send(struct pool *pool, char *s, ssize_t len)
 			break;
 		case SEND_SELECTFAIL:
 			applog(LOG_DEBUG, "Write select failed on pool %d sock", pool->pool_no);
+			suspend_stratum(pool);
 			break;
 		case SEND_SENDFAIL:
-			applog(LOG_DEBUG, "Failed to curl_easy_send in stratum_send");
+			applog(LOG_DEBUG, "Failed to send in stratum_send");
+			suspend_stratum(pool);
 			break;
 		case SEND_INACTIVE:
 			applog(LOG_DEBUG, "Stratum send failed due to no pool stratum_active");
@@ -1079,7 +1063,10 @@ static void clear_sock(struct pool *pool)
 
 	mutex_lock(&pool->stratum_lock);
 	do {
-		n = recv(pool->sock, pool->sockbuf, RECVSIZE, 0);
+		if (pool->sock)
+			n = recv(pool->sock, pool->sockbuf, RECVSIZE, 0);
+		else
+			n = 0;
 	} while (n > 0);
 	mutex_unlock(&pool->stratum_lock);
 
@@ -1132,11 +1119,13 @@ char *recv_line(struct pool *pool)
 			n = recv(pool->sock, s, RECVSIZE, 0);
 			if (!n) {
 				applog(LOG_DEBUG, "Socket closed waiting in recv_line");
+				suspend_stratum(pool);
 				break;
 			}
 			if (n < 0) {
 				if (!sock_blocks() || !socket_full(pool, false)) {
 					applog(LOG_DEBUG, "Failed to recv sock in recv_line");
+					suspend_stratum(pool);
 					break;
 				}
 			} else {
@@ -1870,3 +1859,74 @@ void RenameThread(const char* name)
 #endif
 }
 
+/* cgminer specific wrappers for true unnamed semaphore usage on platforms
+ * that support them and for apple which does not. We use a single byte across
+ * a pipe to emulate semaphore behaviour there. */
+#ifdef __APPLE__
+void cgsem_init(cgsem_t *cgsem)
+{
+	int flags, fd, i;
+
+	if (pipe(cgsem->pipefd) == -1)
+		quit(1, "Failed pipe in cgsem_init");
+
+	/* Make the pipes FD_CLOEXEC to allow them to close should we call
+	 * execv on restart. */
+	for (i = 0; i < 2; i++) {
+		fd = cgsem->pipefd[i];
+		flags = fcntl(fd, F_GETFD, 0);
+		flags |= FD_CLOEXEC;
+		if (fcntl(fd, F_SETFD, flags) == -1)
+			quit(1, "Failed to fcntl in cgsem_init");
+	}
+}
+
+void cgsem_post(cgsem_t *cgsem)
+{
+	const char buf = 1;
+	int ret;
+
+	ret = write(cgsem->pipefd[1], &buf, 1);
+	if (unlikely(ret == 0))
+		applog(LOG_WARNING, "Failed to write in cgsem_post");
+}
+
+void cgsem_wait(cgsem_t *cgsem)
+{
+	char buf;
+	int ret;
+
+	ret = read(cgsem->pipefd[0], &buf, 1);
+	if (unlikely(ret == 0))
+		applog(LOG_WARNING, "Failed to read in cgsem_wait");
+}
+
+void cgsem_destroy(cgsem_t *cgsem)
+{
+	close(cgsem->pipefd[1]);
+	close(cgsem->pipefd[0]);
+}
+#else
+void cgsem_init(cgsem_t *cgsem)
+{
+	if (sem_init(cgsem, 0, 0))
+		quit(1, "Failed to sem_init in cgsem_init");
+}
+
+void cgsem_post(cgsem_t *cgsem)
+{
+	if (unlikely(sem_post(cgsem)))
+		quit(1, "Failed to sem_post in cgsem_post");
+}
+
+void cgsem_wait(cgsem_t *cgsem)
+{
+	if (unlikely(sem_wait(cgsem)))
+		quit(1, "Failed to sem_wait in cgsem_wait");
+}
+
+void cgsem_destroy(cgsem_t *cgsem)
+{
+	sem_destroy(cgsem);
+}
+#endif
